import random
import sys
import pygame


# TODO push to pypi

class LifeGame:

    def __init__(self, screen_width=800, screen_height=600, cell_size=10, alive_color=(0, 255, 255),
                 dead_color=(0, 0, 0)):
        """
        Initialize grid. set default game state, initialize screen

        :param screen_width: Game window width
        :param screen_height: Game window height
        :param cell_size: Diameter of circles.
        :param alive_color: RGB tuple e.g. (255, 255, 255) for cells
        :param dead_color: RGB tuple e.g. (255, 255, 255)
        :param max_fps: Framerate cap to limit game speed
        """

        pygame.init()
        self.screen_width = screen_width
        self.screen_height = screen_height
        self.cell_size = cell_size
        self.alive_color = alive_color
        self.dead_color = dead_color
        self.grid_count = (self.screen_width*self.screen_height)/100

        self.over_pop = 3
        self.under_pop = 2
        self.from_pop = 2
        self.to_pop = 3
        self.cometo_pop = 3

        self.screen = pygame.display.set_mode((self.screen_width, self.screen_height))
        self.clear_screen()
        pygame.display.flip()

        self.last_update_completed = 0
        self.frame_num = 0
        self.old_frame_num = 0
        self.max_fps = 45
        self.desired_milliseconds_between_updates = (1.0 / self.max_fps) * 10000

        self.active_grid = 0
        self.num_cols = int(self.screen_width / self.cell_size)
        self.num_rows = int(self.screen_height / self.cell_size)
        self.grids = []
        self.init_grids()
        self.set_grid()

        self.step = False
        self.paused = False
        self.game_over = False

    def init_grids(self):
        """
        Create and stores the default active and inactive grid

        :return: None
        """

        def create_grid():
            """
            Generate en empty 2d grid
            :return:
            """
            rows = []
            for row_num in range(self.num_rows):
                list_of_columns = [0] * self.num_cols
                rows.append(list_of_columns)
            return rows

        self.grids.append(create_grid())
        self.grids.append(create_grid())

    def set_grid(self, value=None, grid=0):
        """
        Set an entire grid at once. Set to a single value or random 0/1.

        Examples:
          # set_grid(0) # all dead
          # set_grid(1) # all alive
          # set_grid() # random
          # set_grid(None) # random

        :param grid: Index of grid, for active/inactive (0 or 1)
        :param value: Value to set the cell to (0 or 1)
        :return:
        """
        for r in range(self.num_rows):
            for c in range(self.num_cols):
                if value is None:
                    cell_value = random.randint(0, 1)
                else:
                    cell_value = value
                self.grids[grid][r][c] = cell_value

    def draw_grid(self):
        """
        Given the grid and cell states, draw the cells on the screen

        :return:
        """
        self.clear_screen()
        for c in range(self.num_cols):
            for r in range(self.num_rows):
                if self.grids[self.active_grid][r][c] == 1:
                    color = self.alive_color
                else:
                    color = self.dead_color
                pygame.draw.circle(self.screen,
                                   color,
                                   (int(c * self.cell_size + (self.cell_size / 2)),
                                    int(r * self.cell_size + (self.cell_size / 2))),
                                   int(self.cell_size / 2),
                                   0)
        pygame.display.flip()

    def clear_screen(self):
        """
        Fill whole screen with dead color

        :return:
        """
        self.screen.fill(self.dead_color)

    def get_cell(self, row_num, col_num):
        """
        Get the alive/dead (0/1) state of a specific cell in active grid
        
        :param row_num: 
        :param col_num: 
        :return: 0 or 1 depending on state of cell. Defaults to 0 (dead)
        """
        try:
            cell_value = self.grids[self.active_grid][row_num][col_num]
        except:
            cell_value = 0
        return cell_value

    def check_cell_neighbors(self, row_index, col_index):
        """
        Get the number of alive neighbor cells, and determine the state of the cell
        for the next generation, Determine whetjer it lives, dies, survives, or is born.

        :param row_index: Row number of the cell to check
        :param col_index: Column number of the cell to check
        :return: The state the cell should be in next generation (0 or 1)
        """
        num_alive_neighbors = 0
        num_alive_neighbors += self.get_cell(row_index - 1, col_index - 1)
        num_alive_neighbors += self.get_cell(row_index - 1, col_index)
        num_alive_neighbors += self.get_cell(row_index - 1, col_index + 1)
        num_alive_neighbors += self.get_cell(row_index, col_index - 1)
        num_alive_neighbors += self.get_cell(row_index, col_index + 1)
        num_alive_neighbors += self.get_cell(row_index + 1, col_index - 1)
        num_alive_neighbors += self.get_cell(row_index + 1, col_index)
        num_alive_neighbors += self.get_cell(row_index + 1, col_index + 1)

        # Rules for life and death
        if self.grids[self.active_grid][row_index][col_index] == 1:  # alive
            if num_alive_neighbors > self.over_pop:  # Overpopulation
                return 0
            if num_alive_neighbors < self.under_pop:  # Underpopulation
                return 0
            if num_alive_neighbors == self.from_pop or num_alive_neighbors == self.to_pop:
                return 1
        elif self.grids[self.active_grid][row_index][col_index] == 0:  # dead
            if num_alive_neighbors == self.cometo_pop:
                return 1  # come to life

        return self.grids[self.active_grid][row_index][col_index]

    def update_generation(self):
        """
        Inspect current generation state, prepare next generation

        :return:
        """
        self.set_grid(0, self.inactive_grid())
        for r in range(self.num_rows - 1):
            for c in range(self.num_cols - 1):
                if self.frame_num > self.old_frame_num + self.grid_count:  # old_frame_num + grid_count = 1 generation
                    if self.step:  # (step == True) = Frame by Frame is ON
                        input("Press Enter for next generation...")
                    self.old_frame_num = self.frame_num
                    print('Generation: ' + str(int(self.frame_num/self.grid_count)))
                next_gen_state = self.check_cell_neighbors(r, c)
                self.grids[self.inactive_grid()][r][c] = next_gen_state
                self.frame_num += 1
        self.active_grid = self.inactive_grid()

    def inactive_grid(self):
        """
        Simple helper function to get the index of the inactive grid
        If active grid is 0 will return 1 and vice-versa

        :return:
        """
        return (self.active_grid + 1) % 2

    def handle_events(self):
        """
        Handle any keypresses
        s - Start/Stop (pause)
        q - Quit
        r - Randomize grid
        l - Less alive
        m - More alive
        f - Faster
        w - sloWer
        b - frame By frame (ON/OFF)

        :return:
        """
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                if event.unicode == 's':
                    if self.paused:
                        self.paused = False
                    else:
                        self.paused = True
                elif event.unicode == 'r':
                    self.active_grid = 0
                    self.set_grid(None, self.active_grid)  # randomize
                    self.set_grid(0, self.inactive_grid())  # set to 0
                    self.draw_grid()
                elif event.unicode == 'q':
                    self.game_over = True
                elif event.unicode == 'l':
                    self.over_pop += 1
                    self.under_pop += 1
                    self.from_pop += 1
                    self.to_pop += 1
                    self.cometo_pop += 1
                    print('Death if more than: ' + str(self.over_pop)
                          + '   Death if less than: ' + str(self.under_pop)
                          + '   Alive from: ' + str(self.from_pop)
                          + '   Alive to: ' + str(self.to_pop)
                          + '   Come alive at: ' + str(self.cometo_pop))
                elif event.unicode == 'm':
                    self.over_pop -= 1
                    self.under_pop -= 1
                    self.from_pop -= 1
                    self.to_pop -= 1
                    self.cometo_pop -= 1
                    print('Death if more than: ' + str(self.over_pop)
                          + '   Death if less than: ' + str(self.under_pop)
                          + '   Alive from: ' + str(self.from_pop)
                          + '   Alive to: ' + str(self.to_pop)
                          + '   Come alive at: ' + str(self.cometo_pop))
                elif event.unicode == 'f':
                    self.max_fps += 1.0
                    self.desired_milliseconds_between_updates = (1.0 / self.max_fps) * 10000
                    print('FPS: ' + str(self.max_fps))
                elif event.unicode == 'w':
                    if self.max_fps > 1:
                        self.max_fps -= 1.0
                        self.desired_milliseconds_between_updates = (1.0 / self.max_fps) * 10000
                        print('FPS: ' + str(self.max_fps))
                    else:
                        self.max_fps = 1.0
                        self.desired_milliseconds_between_updates = (1.0 / self.max_fps) * 10000
                        print('FPS: ' + str(self.max_fps))
                elif event.unicode == 'b':
                    if self.step:
                        self.step = False
                    else:
                        self.step = True
            if event.type == pygame.QUIT:
                sys.exit()

    def run(self):
        """
        Kick off the game and loop forever until quit state

        :return:
        """
        while True:
            if self.game_over:
                return

            self.handle_events()
            if self.paused:
                continue

            self.update_generation()
            self.draw_grid()
            self.cap_frame_rate()

    def cap_frame_rate(self):
        """
        If game is running too fast and updating game too frequently.
        just wait to maintain stable framerate

        :return:
        """
        now = pygame.time.get_ticks()
        milliseconds_since_last_update = now - self.last_update_completed
        time_to_sleep = self.desired_milliseconds_between_updates - milliseconds_since_last_update
        if time_to_sleep > 0:
            pygame.time.delay(int(time_to_sleep))
        self.last_update_completed = now
